/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is JavaScript Engine testing utilities.
 *
 * The Initial Developer of the Original Code is
 * Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Bryant Chen
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var gTestfile = '10.6.1-01.js';
//-----------------------------------------------------------------------------
var BUGNUMBER = 290774;
var summary = 'activation object never delegates to Object.prototype';
var actual = '';
var expect = '';

printBugNumber(BUGNUMBER);
printStatus (summary);

var toStringResult;
var evalResult;
var watchResult;
var parseIntResult;

var eval = 'fooEval';
var watch = undefined;
var parseInt = 'fooParseInt';


function toString()
{
  return 'fooString';
}

function normal()
{
  toStringResult = toString;
  evalResult = eval;
  watchResult = watch;
  parseIntResult = parseInt;
}

function outerinnervar()
{
  toStringResult = toString;
  evalResult = eval;
  watchResult = watch;
  parseIntResult = parseInt;
  function inner()
  {
    // addition of any statement
    // which accesses a variable
    // from the outer scope causes the bug
    printStatus(toString);
  }
}

expect = true;

printStatus('normal');
printStatus('======');
normal();

printStatus('toStringResult ' + toStringResult);
printStatus('toString ' + toString);
actual = ((toStringResult + '') == (toString + ''));
reportCompare(expect, actual, inSection(1));

printStatus('evalResult ' + evalResult);
printStatus('eval ' + eval);
actual = ((evalResult + '') == (eval + ''));
reportCompare(expect, actual, inSection(2));

printStatus('watchResult ' + watchResult);
printStatus('watch ' + watch);
actual = ((watchResult + '') == (watch + ''));
reportCompare(expect, actual, inSection(3));

printStatus('parseIntResult ' + parseIntResult);
printStatus('parseInt ' + parseInt);
actual = ((parseIntResult + '') == (parseInt + ''));
reportCompare(expect, actual, inSection(4));

printStatus('outerinner');
printStatus('==========');

outerinnervar();

printStatus('toStringResult ' + toStringResult);
printStatus('toString ' + toString);
actual = ((toStringResult + '') == (toString + ''));
reportCompare(expect, actual, inSection(5));


printStatus('evalResult ' + evalResult);
printStatus('eval ' + eval);
actual = ((evalResult + '') == (eval + ''));
reportCompare(expect, actual, inSection(6));

printStatus('watchResult ' + watchResult);
printStatus('watch ' + watch);
actual = ((watchResult + '') == (watch + ''));
reportCompare(expect, actual, inSection(7));

printStatus('parseIntResult ' + parseIntResult);
printStatus('parseInt ' + parseInt);
actual = ((parseIntResult + '') == (parseInt + ''));
reportCompare(expect, actual, inSection(8));
